package egs.wolves;

import static egs.wolves.Move.*;

import java.util.Arrays;
import java.util.EnumSet;
import java.util.Set;
import static util.Util.asSet;

/**
 * A state descriptor for our world. This class follows the value pattern.
 * To construct one, use the constant INITIAL_WORLD.
 * To change the Grid, call applyMove().
 *
 * You can construct one via given coordinates for the wolves and the
 * rabbit.
 */
class Grid implements java.io.Serializable {
  static final int WIDTH = 4;
  static final int HEIGHT = 3;

  private static final int[] DIMENSIONS = new int[]{WIDTH,HEIGHT};
  private static final boolean[] DIMENSION_WRAPS = {true, false};
  public static final int X_AXIS = 0, Y_AXIS = 1;

  static final Grid INITIAL = new Grid(new int[]{0,0},
                                         new int[]{0,0},
                                         new int[]{WIDTH - 1,
                                                    HEIGHT - 1});

  static final Grid[] ALL;
  static final Set<Grid> ALL_POSSIBLE;
  static{
    int n = WIDTH * HEIGHT;
    Grid[] sts = new Grid[n * n * n];
    int s = 0;
    for (int w1x = 0; w1x < WIDTH; w1x++) {
      for (int w1y = 0; w1y < HEIGHT; w1y++) {
        for (int w2x = 0; w2x < WIDTH; w2x++) {
          for (int w2y = 0; w2y < HEIGHT; w2y++) {
            for (int rx = 0; rx < WIDTH; rx++) {
              for (int ry = 0; ry < HEIGHT; ry++) {
                sts[s++] = new Grid(w1x, w1y, w2x, w2y, rx, ry);
              }
            }
          }
        }
      }
    }
    ALL = sts;
    ALL_POSSIBLE = asSet(ALL);
  }



  // x and y positions in an array, *do not mutate*
  // reconstruct a new object
  protected final int[] w1pos, w2pos, rPos;

  protected Grid(int w1x, int w1y,
                int w2x, int w2y,
                int rx, int ry){
    this(new int[]{w1x, w1y}, new int[]{w2x, w2y}, new int[]{rx, ry});
  }

  // do not call directly, used internally
  private Grid(int[] w1, int[] w2, int[] r) {
    w1pos = w1;
    w2pos = w2;
    rPos = r;
  }


  ///////////////////////////////////////////////////////
  // Methods to move various critters around the world
  //
  Grid moveWolf1(Move m){
    return applyMove(this,m,null,null);
  }
  Grid moveWolf2(Move m){
    return applyMove(this,null,m,null);
  }
  Grid moveRabbit(Move m){
    return applyMove(this,null,null,m);
  }

  Set<Move> validMoves(int aix){
    final int[] pos;
    if (aix == 0)
      pos = w1pos;
    else if (aix == 1)
      pos = w2pos;
    else if (aix == 2)
      pos = rPos;
    else
      throw new IllegalArgumentException();

    Set<Move> mvSet = EnumSet.of(Move.STAY);
    for (Move m : Move.SET) {
      if (!Arrays.equals(pos, moveFrom(pos, m))) {
        mvSet.add(m);
      }
    }
    return mvSet;
  }

  public static void main(String[] args){
    System.out.println(Grid.INITIAL);
    System.out.println(Grid.INITIAL.validMoves(0));
    System.out.println(Grid.INITIAL.validMoves(1));
    System.out.println(Grid.INITIAL.validMoves(2));
    System.out.println(Grid.INITIAL.validMoves(3));
  }

  /**
   * Called to change the Grid, given some moves; the <tt>null</tt>
   * reference can be passed in for any animal's move that you want to
   * ignore (same as using Move.STAY).
   *
   * E.g.
   *   applyMove(w, Move.LEFT, null, Move.STAY)
   *    moves wolf 1 left, leaves wolf 2, and the rabbit as well.
   */
  Grid applyMove(Grid g, Move w1, Move w2, Move r){
    return new Grid(moveFrom(g.w1pos, w1),
                      moveFrom(g.w2pos, w2),
                      moveFrom(g.rPos, r));
  }

  private int[] moveFrom(int[] xy, Move m){
    if (m == null || m == Move.STAY)
      return xy;

    int[] xy1 = new int[]{xy[0], xy[1]};
    switch(m) {
    case LEFT:
      xy1[0] -= 1;
      break;
    case RIGHT:
      xy1[0] += 1;
      break;
    case UP:
      xy1[1] -= 1;
      break;
    case DOWN:
      xy1[1] += 1;
      break;
    default: throw new RuntimeException("unimplemented case "+m);
    }
    wrapClamp(xy1);
    return xy1;
  }

  // permits moves around
  private void wrapClamp(int[] xy){
    if (DIMENSION_WRAPS[X_AXIS]) {
      xy[X_AXIS] = (xy[X_AXIS] + WIDTH) % WIDTH;
    } else {
      xy[X_AXIS] = Math.min(WIDTH - 1, Math.max(0, xy[X_AXIS]));
    }
    if (DIMENSION_WRAPS[Y_AXIS]) {
      xy[Y_AXIS] = (xy[Y_AXIS] + HEIGHT) % HEIGHT;
    } else {
      xy[Y_AXIS] = Math.min(HEIGHT - 1, Math.max(0, xy[Y_AXIS]));
    }
  }

  /**
   * This method returns the distance between two points.
   * However, it also adjusts for wrapping and such. Hence, in the
   * map row
   *  r.....1
   * The r is one unit from the 1 since a move left wraps.
   */
  static int dist(int[] xy0, int[] xy1) {
    return manhattanDistance(xy0, xy1);
  }

  static int manhattanDistance(int[] xy0, int[] xy1) {
    return Math.abs(shortestDirectionTo(xy0,xy1,X_AXIS)) +
            Math.abs(shortestDirectionTo(xy0,xy1,Y_AXIS));
  }

/*
  // not exactly what we want
  //static double euclideanDistance(int[] xy0, int[] xy1) {
  //  return Math.hypot(distWrap(xy0,xy1,X_AXIS), xy1[Y_AXIS] - xy0[Y_AXIS]);
  //}

//  public static void main(String[] args){
    // Grid w = new Grid(0,0,0,0,0,0);
    int[] w1 = new int[]{0,0},
          w2 = new int[]{3,2};
////    int[] r = new int[]{WIDTH-1, HEIGHT-1};
    int[] r = new int[]{0,2};
    Grid g = new Grid(w1,w2,r);
//
//    System.out.printf("         W1     W2%n");
//    System.out.printf("dx %7s %7s%n", shortestDirectionTo(w1,r,X_AXIS),shortestDirectionTo(w2,r,X_AXIS));
//    System.out.printf("dy %7s %7s%n", shortestDirectionTo(w1,r,Y_AXIS),shortestDirectionTo(w2,r,Y_AXIS));
//    System.out.printf("d  %7s %7s%n", manhattanDistance(w1,r),manhattanDistance(w2,r));
//    System.out.printf("mv %7s %7s%n", w.bestMove1(), w.bestMove2());
    System.out.println(g);
//    System.out.println();
//    System.out.println(w.bestMove2());
  }
*/

  // wolf1's best move
  public Move bestMove1(){
    return bestMove(w1pos);
  }

  public Move bestMove2(){
    return bestMove(w2pos);
  }

  private Move bestMove(int[] from){
    int dx = shortestDirectionTo(from, rPos, X_AXIS),
        dy = shortestDirectionTo(from, rPos, Y_AXIS);

    if (Math.abs(dx) >= Math.abs(dy)) {
      return dx > 0 ? Move.RIGHT :
              dx < 0 ? Move.LEFT : Move.STAY;
    } else {
      return dy > 0 ? Move.DOWN : Move.UP;
      // dy == 0 is impossible
    }
  }

  // computes the distance vector based on the shortest distance from a point
  // to another point.
  private static int shortestDirectionTo(int[] xyFrom, int[] xyTo, int axis) {
    if (DIMENSION_WRAPS[axis]) {
      int from = xyFrom[axis],
            to = xyTo[axis];

      int directDir = to - from;
      int wrapDir;
      if (from <= to) {
        wrapDir = -(from + DIMENSIONS[axis] - to);
      } else {
        wrapDir = (to + DIMENSIONS[axis] - from);
      }
      return (Math.abs(directDir) <= Math.abs(wrapDir)) ? directDir : wrapDir;

    } else { // no wrap
      return xyTo[axis] - xyFrom[axis];
    }
  }

  boolean rabbitCaught() {
    return Arrays.equals(w1pos, rPos) ||
            Arrays.equals(w2pos, rPos);
  }

  int closestWolfDist(){
    return Math.min(dist(w1pos,rPos), dist(w2pos,rPos));
  }



  @Override
  public int hashCode(){
    return Arrays.hashCode(w1pos) ^
            Arrays.hashCode(w2pos) ^
             Arrays.hashCode(rPos);
  }

  @Override
  public boolean equals(Object o){
    Grid w = (Grid)o;
    return Arrays.equals(w1pos, w.w1pos) &&
            Arrays.equals(w2pos, w.w2pos) &&
             Arrays.equals(rPos, w.rPos);
  }

  public String toShortString(){
    return String.format("<%2d,%2d>, <%2d,%2d>, <%2d,%2d>",
                          w1pos[0], w1pos[1],
                            w2pos[0], w2pos[1],
                              rPos[0], rPos[1]);
  }

  /**
   * Prints a textual representation of the map.
   *  . indicates a clear square
   *  r indicates the rabbit (alone)
   *  1 indicates the first wolf (alone)
   *  2 indicates the second wolf (alone)
   *  3 indicates both wolves on the same square
   *  x indicates one wolf with the rabbit (captured).
   *  X indicates both wolves with the rabbit (captured).
   */
  public String toString(){
    int[] xy = new int[2];
    StringBuilder sb = new StringBuilder((WIDTH + 1) * HEIGHT);
    sb.append('+');
    for (int x=0; x < WIDTH; x++)
      if (DIMENSION_WRAPS[Y_AXIS])
        sb.append('^');
      else
        sb.append('-');
    sb.append('+').append('\n');

    for (int y=0; y < HEIGHT; y++) {
      if (DIMENSION_WRAPS[X_AXIS])
        sb.append('<');
      else
        sb.append('|');

      xy[1] = y;
      for (int x=0; x < WIDTH; x++) {
        xy[0] = x;
        if (Arrays.equals(xy, w1pos) &&
              Arrays.equals(xy, w2pos) &&
              Arrays.equals(xy, rPos)) {
          sb.append('X');
        } else if ((Arrays.equals(xy, w1pos) &&
                        Arrays.equals(xy, rPos)) ||
                        Arrays.equals(xy, w2pos) &&
                        Arrays.equals(xy, rPos)) {
          sb.append('x');
        } else if (Arrays.equals(xy, w1pos) && Arrays.equals(xy, w2pos)) {
          sb.append('3');
        } else if (Arrays.equals(xy, rPos)) {
          sb.append('r');
        } else if (Arrays.equals(xy, w1pos)) {
          sb.append('1');
        } else if (Arrays.equals(xy, w2pos)) {
          sb.append('2');
        } else {
          sb.append('.');
        }
      }
      if (DIMENSION_WRAPS[X_AXIS])
        sb.append('>');
      else
        sb.append('|');

      sb.append('\n');
    }

    sb.append('+');
    for (int x=0; x < WIDTH; x++)
      if (DIMENSION_WRAPS[Y_AXIS])
        sb.append('v');
      else
        sb.append('-');
    sb.append('+').append('\n');


    return sb.toString();
  }
}
